// CMemPatch
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2011 Dominik Friedrichs

#include "CMemPatch.h"


CMemPatch::CMemPatch(void)
{
  PatchAddress = NULL;

  PatchData = NULL;
  PatchLength = 0;

  MemoryBackup = NULL;
  MemoryPatched = false;

  return;
}


CMemPatch::~CMemPatch(void)
{
  FreePatchMemory();
}


bool CMemPatch::PatchMemory(const char *Patch,
                            void *Address, ...)
{
  if(MemoryPatched
  || !Patch || !Address)
    return(false);

  PatchAddress = NULL;
  FreePatchMemory();

  va_list ap;
  va_start(ap, Address);

  int patchlen = HexPatchToBinary(Patch,
                                  PatchData,
                                  Address, ap);

  va_end(ap);

  try
  {
    if(patchlen <= 0)
    // error or zero length
      throw(-1);

    PatchAddress = (unsigned char *)Address;
    PatchLength = patchlen;

    // backup memory
    MemoryBackup = new unsigned char[PatchLength];

    if(!MemReadRaw(PatchAddress, MemoryBackup, PatchLength))
      throw(-1);

    // patch memory
    if(!MemWriteRaw(PatchAddress, PatchData, PatchLength))
      throw(-1);

    MemoryPatched = true;
  }
  catch(int)
  {
    FreePatchMemory();
    return(false);
  }

  return(true);
}


bool CMemPatch::RestoreMemory(void)
{
  if(!MemoryPatched)
    return(false);

  // restore backup
  if(!MemWriteRaw(PatchAddress, MemoryBackup, PatchLength))
    return(false);

  MemoryPatched = false;
  return(true);
}


void CMemPatch::FreePatchMemory(void)
{
  delete [] PatchData;
  delete [] MemoryBackup;

  PatchData = NULL;
  MemoryBackup = NULL;

  PatchLength = 0;
  return;
}


bool CMemPatch::MemReadRaw(const void *readaddr,
                           void *buffer,
                           size_t length)
{
  HANDLE process = GetCurrentProcess();

  DWORD nativeprot;
  DWORD memcpyprot = PAGE_EXECUTE_READ;

  if(!process
  || !VirtualProtect((LPVOID)readaddr, length, memcpyprot, &nativeprot)
  || !FlushInstructionCache(process, readaddr, length))
  // couldnt get safe memory access
    return(false);

  memcpy(buffer, readaddr, length);

  // restore native protection
  DWORD tempprot;
  VirtualProtect((LPVOID)readaddr, length, nativeprot, &tempprot);
  return(true);
}


bool CMemPatch::MemWriteRaw(void *writeaddr,
                            const void *buffer,
                            size_t length)
{
  HANDLE process = GetCurrentProcess();

  DWORD nativeprot;
  DWORD memcpyprot = PAGE_EXECUTE_READWRITE;

  if(!process
  || !VirtualProtect(writeaddr, length, memcpyprot, &nativeprot)
  || !FlushInstructionCache(process, writeaddr, length))
  // couldnt get safe memory access
    return(false);

  memcpy(writeaddr, buffer, length);

  // restore native protection
  DWORD tempprot;
  VirtualProtect(writeaddr, length, nativeprot, &tempprot);
  return(true);
}


int CMemPatch::HexPatchToBinary(const char *str,
                                unsigned char *&data,
                                void *baseaddr, va_list ap)
{
  if(!str || !&data
  || !baseaddr)
    return(-1);

  unsigned int len = (strlen(str)+1)/3;

  if(!len)
    return(-1);

  unsigned int byte;
  data = new unsigned char[len];

  for(byte = 0; *str;)
  {
    if(byte > 0)
    {
      if(byte >= len
      || *str != '\x20')
      // overflow or not whitespace
      {
        byte = 0;
        break;
      }

      // skip spaces inbetween
      str++;
    }

    if(!strncmp(str, "## ## ## ##", 11))
    // relative 32bit address placeholder
    {
      unsigned int reladdr = (unsigned char *)va_arg(ap, void *)
                           -((unsigned char *)baseaddr + byte + 4);
      memcpy(&data[byte], &reladdr, 4);

      byte += 4;
      str += 11;
      continue;
    }

    char hexp[3];
    strncpy(hexp, str, 2);
    hexp[2] = '\0';

    if(!IsHexChar(hexp[0])
    || !IsHexChar(hexp[1]))
    // invalid byte
      break;

    data[byte] = (unsigned char)strtoul(hexp, NULL, 16);

    byte += 1;
    str += 2;
  }

  if(byte != len)
  // loop terminated abnormally
    return(-1);

  return(byte);
}


bool CMemPatch::IsHexChar(char c)
{
  if((c >= '0' && c <= '9')
  || (c >= 'a' && c <= 'f')
  || (c >= 'A' && c <= 'F'))
    return(true);

  return(false);
}
